from setuptools import setup

setup(
    name='giveme',
    version='0.3.1',
    description='Tiny dependency injection framework',
    author='Steinthor Palsson',
    author_email='steini90@gmail.com',
    url='https://github.com/steinitzu/giveme',
    license='MIT',
    install_requires=[],
    packages=['giveme']
)
